import 'dart:core';
import 'dart:math';

import 'package:collection/collection.dart';
import 'package:meta/meta.dart';

import '../idisposable.dart';
import 'inotify_property_changed.dart';
import 'notify_property_changed.dart';

typedef Transformer<TItem, TTransformedItem> = TTransformedItem Function(
    TItem item);

abstract class ItemAddedOrRemovedEvent extends PropertyChangedEvent {
  // Properties

  /// The item that has been added or removed
  final Object item;

  // The index of the item property;
  final int index;

  // Methods
  ItemAddedOrRemovedEvent(
      Object sender, String propertyName, this.item, this.index)
      : assert(item != null),
        assert(index != null && index >= 0),
        super(sender, propertyName);
}

class ItemAddedEvent extends ItemAddedOrRemovedEvent {
  // Properties
  static const String operationName = "add";

  // Methods
  ItemAddedEvent(Object sender, Object item, int index)
      : super(sender, operationName, item, index);
}

class ItemRemovedEvent extends ItemAddedOrRemovedEvent {
  // Properties
  static const String operationName = "remove";

  // Methods
  ItemRemovedEvent(Object sender, Object item, int index)
      : super(sender, operationName, item, index);
}

/// When ever the list changes a notification is send.
class ObservableList<E> extends DelegatingList<E> with NotifyPropertyChanged {
  // Properties
  /// Property name to indicate that something has been changed;
  static const String itemsChangedPropertyName = "items";

  final List<E> _internal;
  final bool _disposeItems;
  bool _suppressChangeNotification = false;

  @override
  set first(E value) =>
      _updateValue(_internal.first, value, (v) => _internal.first = v);

  @override
  E get last => _internal.last;

  @override
  set last(E value) =>
      _updateValue(_internal.last, value, (v) => _internal.last = v);

  @override
  int get length => _internal.length;
  @override
  set length(int value) =>
      _updateValue(_internal.length, value, (v) => _internal.length = v);

  @override
  E operator [](int index) => _internal[index];

  @override
  void operator []=(int index, E value) =>
      _updateValue(_internal[index], value, (v) => _internal[index] = v);

  // Methods

  ObservableList([bool disposeItems = true]) : this._(<E>[], disposeItems);

  ObservableList._(internal, bool disposeItems)
      : _internal = internal,
        _disposeItems = disposeItems,
        super(internal);

  @override
  void dispose() {
    _internal.forEach((item) => disposeItem(item));
    _internal.clear();
    super.dispose();
  }

  void disposeItem(E item) {
    (item as IDisposable)?.dispose();
  }

  bool _updateValue<TPropertyType>(TPropertyType currentValue,
      TPropertyType newValue, SetValue<TPropertyType> setNewValue) {
    return updateValue(
        itemsChangedPropertyName, currentValue, newValue, setNewValue);
  }

  void _notifyItemsChanged() {
    notifyPropertyChanged(itemsChangedPropertyName);
  }

  @protected
  void _notifiyItemAdded(E item) {
    final event = ItemAddedEvent(this, item, indexOf(item));
    propertyChanged.add(event);
  }

  @protected
  void _notifItemRemoved(E item, int index) {
    final event = ItemRemovedEvent(this, item, index);
    propertyChanged.add(event);
  }

  @override
  void notifyPropertyChanged(String propertyName) {
    if (_suppressChangeNotification) {
      return;
    }
    propertyChanged.add(PropertyChangedEvent(this, propertyName));
  }

  @override
  void add(E value) {
    _internal.add(value);
    _notifyItemsChanged();
    _notifiyItemAdded(value);
  }

  @override
  void addAll(Iterable<E> iterable) {
    _internal.addAll(iterable);
    _notifyItemsChanged();
  }

  void addTransformed<TItem>(
      Iterable<TItem> iterable, Transformer<TItem, E> transformer) {
    _suppressChangeNotification = true;
    try {
      iterable.forEach((item) => add(transformer(item)));
    } finally {
      _suppressChangeNotification = false;
    }
  }

  @override
  void clear() {
    if (_disposeItems) {
      _internal.forEach((item) => disposeItem(item));
    }
    _internal.clear();
    _notifyItemsChanged();
  }

  @override
  void fillRange(int start, int end, [E fillValue]) {
    _internal.fillRange(start, end, fillValue);
    _notifyItemsChanged();
  }

  @override
  void insert(int index, E element) {
    _internal.insert(index, element);
    _notifyItemsChanged();
    _notifiyItemAdded(element);
  }

  @override
  void insertAll(int index, Iterable<E> iterable) {
    _internal.insertAll(index, iterable);
    _notifyItemsChanged();
  }

  @override
  bool remove(Object value) {
    final idx = _internal.indexOf(value);
    final result = _internal.remove(value);
    _notifyItemsChanged();
    _notifItemRemoved(value, idx);
    return result;
  }

  @override
  E removeAt(int index) {
    var result = _internal.removeAt(index);
    _notifyItemsChanged();
    _notifItemRemoved(result, index);
    return result;
  }

  @override
  E removeLast() {
    final lastIdx = length-1; 
    final result = _internal.removeLast();
    _notifyItemsChanged();
    _notifItemRemoved(result, lastIdx);
    return result;
  }

  @override
  void removeRange(int start, int end) {
    _internal.removeRange(start, end);
    _notifyItemsChanged();
  }

  @override
  void removeWhere(bool Function(E element) test) {
    _internal.removeWhere(test);
    _notifyItemsChanged();
  }

  @override
  void replaceRange(int start, int end, Iterable<E> replacement) {
    _internal.replaceRange(start, end, replacement);
    _notifyItemsChanged();
  }

  @override
  void retainWhere(bool Function(E element) test) {
    _internal.retainWhere(test);
    _notifyItemsChanged();
  }

  @override
  void setAll(int index, Iterable<E> iterable) {
    _internal.setAll(index, iterable);
    _notifyItemsChanged();
  }

  @override
  void setRange(int start, int end, Iterable<E> iterable, [int skipCount = 0]) {
    _internal.setRange(start, end, iterable);
    _notifyItemsChanged();
  }

  @override
  void shuffle([Random random]) {
    _internal.shuffle(random);
    _notifyItemsChanged();
  }

  @override
  void sort([int Function(E a, E b) compare]) {
    _internal.sort(compare);
    _notifyItemsChanged();
  }
}
